@@ -27,7 +27,7 @@ def findPrediction(row):
 	
 	return prediction
 
-def apply(df, config, header, dataset_features, validation_df = None):
+def apply(df, config, header, dataset_features, validation_df = None, process_id = None):
 
 	models = []; alphas = []
 	
@@ -46,6 +46,8 @@ def apply(df, config, header, dataset_features, validation_df = None):
 	final_predictions = pd.DataFrame(np.zeros((df.shape[0], 2)), columns = ['Prediction', 'Actual'])
 	final_predictions['Actual'] = df['Decision']
 	
+	best_epoch_idx = 0; best_epoch_value = 1000000
+	
 	#for i in range(0, num_of_weak_classifier):
 	pbar = tqdm(range(0, num_of_weak_classifier), desc='Adaboosting')
 	for i in pbar:
@@ -59,7 +61,7 @@ def apply(df, config, header, dataset_features, validation_df = None):
 		#print(worksheet)
 		Training.buildDecisionTree(worksheet.drop(columns=['Weight'])
 			, root, file, config, dataset_features
-			, parent_level = 0, leaf_id = 0, parents = 'root')
+			, parent_level = 0, leaf_id = 0, parents = 'root', main_process_id = process_id)
 		
 		#---------------------------------------
 		
@@ -104,35 +106,20 @@ def apply(df, config, header, dataset_features, validation_df = None):
 		
 		mae = (np.abs(final_predictions['Prediction'].apply(functions.sign) - final_predictions['Actual'])/2).sum()/final_predictions.shape[0]
 		#print(mae)
+		
+		if mae < best_epoch_value:
+			best_epoch_value = mae * 1
+			best_epoch_idx = i * 1
+		
 		pbar.set_description("Epoch %d. Loss: %d. Process: " % (i+1, mae))
 	
 	#------------------------------
-	final_predictions['Prediction'] = final_predictions['Prediction'].apply(functions.sign)
-	final_predictions['Absolute_Error'] = np.abs(final_predictions['Actual'] - final_predictions['Prediction'])/2
-	#print(final_predictions)
-	mae = final_predictions['Absolute_Error'].sum() / final_predictions.shape[0]
-	print("Loss (MAE) found ", mae, " with ",num_of_weak_classifier, ' weak classifiers')
 	
-	#------------------------------
-	#evaluate model
+	print("The best epoch is ",best_epoch_idx," with the ",best_epoch_value," MAE score")
 	
-	model = {}
-	model["trees"] = models
-	model["config"] = config
-	model["alphas"] = alphas
+	models = models[0: best_epoch_idx+1]
+	alphas = alphas[0: best_epoch_idx+1]
 	
-	functions.bulk_prediction(df, model)
-	df['Prediction'] = df['Prediction'].astype(str)
-	df['Decision'] = df['Decision'].astype(str)	
-	evaluate.evaluate(df)
-	
-	if isinstance(validation_df, pd.DataFrame):
-		functions.bulk_prediction(validation_df, model)
-		validation_df['Prediction'] = validation_df['Prediction'].astype(str)
-		validation_df['Decision'] = validation_df['Decision'].astype(str)	
-		evaluate.evaluate(validation_df, task = 'validation')
-	
-	del model
 	#------------------------------
 	
 	return models, alphas
