@@ -9,6 +9,8 @@ from chefboost import Chefboost as cb
 
 from tqdm import tqdm
 
+import gc
+
 def findPrediction(row):
 	epoch = row['Epoch']
 	row = row.drop(labels=['Epoch'])
@@ -22,11 +24,12 @@ def findPrediction(row):
 	fp, pathname, description = imp.find_module(moduleName)
 	myrules = imp.load_module(moduleName, fp, pathname, description)
 	
-	prediction = int(myrules.findDecision(params)) 
+	#prediction = int(myrules.findDecision(params)) 
+	prediction = myrules.findDecision(params)
 	
 	return prediction
 
-def regressor(df, config, header, dataset_features, validation_df = None):
+def regressor(df, config, header, dataset_features, validation_df = None, process_id = None):
 	models = []
 	
 	#we will update decisions in every epoch, this will be used to restore
@@ -72,6 +75,8 @@ def regressor(df, config, header, dataset_features, validation_df = None):
 	
 	#------------------------------
 	
+	best_epoch_idx = 0; best_epoch_loss = 1000000
+	
 	pbar = tqdm(range(1, epochs+1), desc='Boosting')
 	
 	#for index in range(1,epochs+1):
@@ -105,11 +110,16 @@ def regressor(df, config, header, dataset_features, validation_df = None):
 		base_df['Boosted_Prediction'] += df['Prediction']
 		
 		loss = (base_df['Boosted_Prediction'] - base_df['Decision']).pow(2).sum()
+		current_loss = loss / num_of_instances #mse
 		
 		if index == 1: 
-			boosted_from = loss / num_of_instances
+			boosted_from = current_loss * 1
 		elif index == epochs:
-			boosted_to = loss / num_of_instances
+			boosted_to = current_loss * 1
+		
+		if current_loss < best_epoch_loss:
+			best_epoch_loss = current_loss * 1
+			best_epoch_idx = index * 1
 		
 		df['Decision'] = int(learning_rate)*(df['Decision'] - df['Prediction'])
 		df = df.drop(columns = ['Epoch', 'Prediction'])
@@ -129,7 +139,7 @@ def regressor(df, config, header, dataset_features, validation_df = None):
 		
 		current_df = df.copy()
 		Training.buildDecisionTree(df,root,file, config, dataset_features
-			, parent_level = 0, leaf_id = 0, parents = 'root')
+			, parent_level = 0, leaf_id = 0, parents = 'root', main_process_id = process_id)
 		
 		#functions.storeRule(json_file," {}]")
 		
@@ -142,30 +152,19 @@ def regressor(df, config, header, dataset_features, validation_df = None):
 		#print("loss: ",loss)
 		pbar.set_description("Epoch %d. Loss: %d. Process: " % (index, loss))
 		
-		#---------------------------------
+		gc.collect()
+		
+	#---------------------------------
 	
-	print("MSE of ",num_of_instances," instances are boosted from ",boosted_from," to ",boosted_to," in ",epochs," epochs")
+	print("The best epoch is ", best_epoch_idx," with ", best_epoch_loss," loss value")
+	models = models[0:best_epoch_idx]
+	config["epochs"] = best_epoch_idx
 	
-	#-------------------------------
-	#evaluate gbm model
-	
-	model = {}
-	model["trees"] = models
-	model["config"] = config
-	
-	functions.bulk_prediction(df, model)
-	df['Decision'] = base_actuals
-	evaluate.evaluate(df)
-	
-	if isinstance(validation_df, pd.DataFrame):
-		functions.bulk_prediction(validation_df, model)
-		evaluate.evaluate(validation_df, task = 'validation')
-	
-	#-------------------------------
+	print("MSE of ",num_of_instances," instances are boosted from ",boosted_from," to ",best_epoch_loss," in ",epochs," epochs")
 	
 	return models
 
-def classifier(df, config, header, dataset_features, validation_df = None):
+def classifier(df, config, header, dataset_features, validation_df = None, process_id = None):
 	
 	models = []
 	
@@ -191,6 +190,9 @@ def classifier(df, config, header, dataset_features, validation_df = None):
 		actual_set[current_class] = np.where(df['Decision'] == current_class, 1, 0)
 	actual_set = actual_set.values #transform it to numpy array
 	
+	best_accuracy_idx = 0; best_accuracy_value = 0
+	accuracies = []
+	
 	#for epoch in range(0, epochs):
 	for epoch in pbar:
 		for i in range(0, len(classes)):
@@ -218,7 +220,7 @@ def classifier(df, config, header, dataset_features, validation_df = None):
 				functions.createFile(json_file, "[\n")
 			
 			Training.buildDecisionTree(temp_df, root, file, config, dataset_features
-				, parent_level = 0, leaf_id = 0, parents = 'root')
+				, parent_level = 0, leaf_id = 0, parents = 'root', main_process_id = process_id)
 				
 			#decision rules created
 			#----------------------------
@@ -239,7 +241,7 @@ def classifier(df, config, header, dataset_features, validation_df = None):
 				
 				actual = temp_df.loc[row]['Decision']
 				prediction = myrules.findDecision(features)
-								
+												
 				predictions.append(prediction)
 					
 			#----------------------------
@@ -288,30 +290,25 @@ def classifier(df, config, header, dataset_features, validation_df = None):
 			if actual == prediction:
 				classified = classified + 1
 		
-		accuracy = str(100 * classified / actual_set.shape[0]) + "%"
+		accuracy = 100 * classified / actual_set.shape[0]
+		accuracies.append(accuracy)
+		
+		if accuracy > best_accuracy_value:
+			best_accuracy_value = accuracy * 1
+			best_accuracy_idx = epoch * 1
 		
 		#----------------------------
 		
 		#print(worksheet.head())
 		#print("round ",epoch+1)
-		pbar.set_description("Epoch %d. Accuracy: %s. Process: " % (epoch+1, accuracy))
-	
-	
+		pbar.set_description("Epoch %d. Accuracy: %d. Process: " % (epoch+1, accuracy))
+		
+		gc.collect()
+		
 	#--------------------------------
-	#evaluate gbm model
 	
-	model = {}
-	model["trees"] = models
-	model["config"] = config
-	model["alphas"] = classes
+	print("The best accuracy got in ",best_accuracy_idx," epoch with the score ", best_accuracy_value)
 	
-	functions.bulk_prediction(df, model)
-	evaluate.evaluate(df)
-	
-	if isinstance(validation_df, pd.DataFrame):
-		functions.bulk_prediction(validation_df, model)
-		evaluate.evaluate(validation_df, task = 'validation')
-	
-	#--------------------------------
+	models = models[0: best_accuracy_idx * len(classes) + len(classes)]
 	
 	return models, classes
